package de.hhhammer.dchat.bot.discord;

import de.hhhammer.dchat.bot.openai.ChatGPTRequestBuilder;
import de.hhhammer.dchat.bot.openai.ChatGPTService;
import de.hhhammer.dchat.bot.openai.ResponseException;
import de.hhhammer.dchat.db.UserDBService;
import de.hhhammer.dchat.db.models.user.UserMessage;
import org.javacord.api.event.message.MessageCreateEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;

public class UserMessageHandler implements MessageHandler {
    private static final Logger logger = LoggerFactory.getLogger(UserMessageHandler.class);
    private final UserDBService userDBService;
    private final ChatGPTService chatGPTService;

    public UserMessageHandler(UserDBService userDBService, ChatGPTService chatGPTService) {
        this.userDBService = userDBService;
        this.chatGPTService = chatGPTService;
    }

    @Override
    public void handle(MessageCreateEvent event) throws ResponseException, IOException, InterruptedException {
        String content = event.getReadableMessageContent();
        var userId = String.valueOf(event.getMessageAuthor().getId());
        var config = this.userDBService.getConfig(userId).get();
        var systemMessage = config.systemMessage();
        var context = this.userDBService.getLastMessages(userId, config.contextLength())
                .stream()
                .map(userMessage -> new ChatGPTRequestBuilder.PreviousInteraction(userMessage.question(), userMessage.answer()))
                .toList();
        var request = new ChatGPTRequestBuilder().contextRequest(context, content, systemMessage);
        var response = this.chatGPTService.submit(request);
        if (response.choices().isEmpty()) {
            event.getMessage().reply("No response available");
            return;
        }
        var answer = response.choices().get(0).message().content();
        logUserMessage(event, content, answer, response.usage().totalTokens());
        event.getChannel().sendMessage(answer);
    }

    @Override
    public boolean isAllowed(MessageCreateEvent event) {
        if (event.getServer().isPresent()) {
            return false;
        }

        var userId = event.getMessageAuthor().getId();
        var config = this.userDBService.getConfig(String.valueOf(userId));
        if (config.isEmpty()) {
            logger.debug("Not allowed with id: " + userId);
            return false;
        }
        return true;
    }

    @Override
    public boolean exceedsRate(MessageCreateEvent event) {
        var userId = String.valueOf(event.getMessageAuthor().getId());
        var config = this.userDBService.getConfig(userId);
        if (config.isEmpty()) {
            logger.error("Missing configuration for userId with id: " + userId);
            return true;
        }
        var rateLimit = config.get().rateLimit();
        var countMessagesInLastMinute = this.userDBService.countMessagesInLastMinute(userId);

        return countMessagesInLastMinute >= rateLimit;
    }

    @Override
    public boolean canHandle(MessageCreateEvent event) {
        return event.isPrivateMessage();
    }

    private void logUserMessage(MessageCreateEvent event, String question, String answer, int tokens) {
        var userId = event.getMessageAuthor().getId();

        var userMessage = new UserMessage.NewUserMessage(String.valueOf(userId), question, answer, tokens);
        this.userDBService.addMessage(userMessage);
    }
}
