package de.hhhammer.dchat.db;

import de.hhhammer.dchat.db.models.user.UserConfig;
import de.hhhammer.dchat.db.models.user.UserMessage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.sql.DataSource;
import java.sql.*;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.Optional;
import java.util.stream.StreamSupport;

public class UserDBService {
    private static final Logger logger = LoggerFactory.getLogger(UserDBService.class);
    private final DataSource dataSource;

    public UserDBService(DataSource dataSource) {
        this.dataSource = dataSource;
    }

    public Optional<UserConfig> getConfig(String userId) {
        var getServerConfig = """
                SELECT * FROM user_configs WHERE user_id = ?
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, userId);
            ResultSet resultSet = pstmt.executeQuery();
            Iterable<UserConfig> iterable = () -> new ResultSetIterator<>(resultSet, new UserConfig.UserConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).findFirst();
        } catch (SQLException e) {
            logger.error("Getting configuration for user with id: " + userId, e);
        } catch (ResultSetIteratorException e) {
            logger.error("Iterating over ServerConfig ResultSet for user with id: " + userId, e);
            return Optional.empty();
        }
        return Optional.empty();
    }

    public Optional<UserConfig> getConfigBy(long id) throws DBException {
        var getServerConfig = """
                SELECT * FROM user_configs WHERE id = ?
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setLong(1, id);
            ResultSet resultSet = pstmt.executeQuery();
            Iterable<UserConfig> iterable = () -> new ResultSetIterator<>(resultSet, new UserConfig.UserConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).findFirst();
        } catch (SQLException e) {
            throw new DBException("Getting configuration id: " + id, e);
        } catch (ResultSetIteratorException e) {
            throw new DBException("Iterating over UserConfig ResultSet with id: " + id, e);
        }
    }

    public List<UserConfig> getAllConfigs() throws DBException {
        var getServerConfig = """
                SELECT * FROM user_configs
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            ResultSet resultSet = pstmt.executeQuery();
            Iterable<UserConfig> iterable = () -> new ResultSetIterator<>(resultSet, new UserConfig.UserConfigResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).toList();
        } catch (SQLException e) {
            throw new DBException("Getting all configurations", e);
        } catch (ResultSetIteratorException e) {
            throw new DBException("Iterating over all UserConfig ResultSet", e);
        }
    }

    public void addConfig(UserConfig.NewUserConfig newUserConfig) throws DBException {
        var getServerConfig = """
                INSERT INTO user_configs (user_id, system_message, context_length, rate_limit) VALUES (?,?,?,?)
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, newUserConfig.userId());
            pstmt.setString(2, newUserConfig.systemMessage());
            pstmt.setInt(3, newUserConfig.contextLength());
            pstmt.setInt(4, newUserConfig.rateLimit());
            int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config added for user with id: " + newUserConfig.userId());
            }
        } catch (SQLException e) {
            throw new DBException("Adding configuration for user with id: " + newUserConfig.userId(), e);
        }
    }

    public void updateConfig(long id, UserConfig.NewUserConfig newUserConfig) throws DBException {
        var getServerConfig = """
                UPDATE user_configs SET system_message = ?, context_length = ?, rate_limit = ?, user_id = ? WHERE id = ?
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, newUserConfig.systemMessage());
            pstmt.setInt(2, newUserConfig.rateLimit());
            pstmt.setLong(3, newUserConfig.contextLength());
            pstmt.setString(4, newUserConfig.userId());
            pstmt.setLong(5, id);
            int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config update with id: " + id);
            }
        } catch (SQLException e) {
            throw new DBException("Updating configuration with id: " + id, e);
        }
    }

    public void deleteConfig(long id) throws DBException {
        var getServerConfig = """
                DELETE FROM user_configs WHERE id = ?
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setLong(1, id);
            int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No config deleted for user with id: " + id);
            }
        } catch (SQLException e) {
            throw new DBException("Deleting configuration with id: " + id, e);
        }
    }

    public int countMessagesInLastMinute(String userId) {
        var getServerConfig = """
                SELECT count(*) FROM user_messages WHERE user_id = ? AND time <= ? and time >= ?
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, userId);
            var now = Instant.now();
            pstmt.setTimestamp(2, Timestamp.from(now));
            pstmt.setTimestamp(3, Timestamp.from(now.minus(1, ChronoUnit.MINUTES)));
            ResultSet resultSet = pstmt.executeQuery();
            resultSet.next();
            return resultSet.getInt(1);
        } catch (SQLException e) {
            logger.error("Getting messages for user with id: " + userId, e);
        } catch (ResultSetIteratorException e) {
            logger.error("Iterating over ServerMessages ResultSet for user with id: " + userId, e);
            return Integer.MAX_VALUE;
        }
        return Integer.MAX_VALUE;
    }

    public void addMessage(UserMessage.NewUserMessage newUserMessage) {
        var getServerConfig = """
                INSERT INTO user_messages (user_id, question, answer, tokens) VALUES (?,?,?,?)
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getServerConfig)
        ) {
            pstmt.setString(1, newUserMessage.userId());
            pstmt.setString(2, newUserMessage.question());
            pstmt.setString(3, newUserMessage.answer());
            pstmt.setInt(4, newUserMessage.tokens());
            int affectedRows = pstmt.executeUpdate();
            if (affectedRows == 0) {
                logger.error("No message added for user with id: " + newUserMessage.userId());
            }
        } catch (SQLException e) {
            logger.error("Adding message to user with id: " + newUserMessage.userId(), e);
        }
    }

    public List<UserMessage> getLastMessages(String userId, int limit) {
        var getLastMessages = """
                SELECT * FROM user_messages WHERE user_id = ? ORDER BY time DESC LIMIT ?
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(getLastMessages)
        ) {
            pstmt.setString(1, userId);
            pstmt.setInt(2, limit);
            ResultSet resultSet = pstmt.executeQuery();
            Iterable<UserMessage> iterable = () -> new ResultSetIterator<>(resultSet, new UserMessage.UserMessageResultSetTransformer());
            return StreamSupport.stream(iterable.spliterator(), false).toList();
        } catch (SQLException e) {
            logger.error("Fetching last messages for user whit id: " + userId, e);
        } catch (ResultSetIteratorException e) {
            logger.error("Iterating over messages ResultSet from user with id: " + userId, e);
        }
        return List.of();
    }

    public long tokensOfLast30Days(String userId) {
        var countTokensOfLast30Days = """
                SELECT sum(tokens) FROM user_messages WHERE user_id = ? AND time < ? AND time >= ?
                """;
        try (Connection con = dataSource.getConnection();
             PreparedStatement pstmt = con.prepareStatement(countTokensOfLast30Days)
        ) {
            pstmt.setString(1, userId);
            var now = Instant.now();
            pstmt.setTimestamp(2, Timestamp.from(now));
            pstmt.setTimestamp(3, Timestamp.from(now.minus(30, ChronoUnit.DAYS)));
            ResultSet resultSet = pstmt.executeQuery();
            resultSet.next();
            return resultSet.getLong(1);
        } catch (SQLException e) {
            logger.error("Counting tokens of the last 30 days from user with id: " + userId, e);
        }
        logger.error("No tokens found for user with id: " + userId);
        return 0;
    }
}
